@@ -1,5 +1,5 @@
 /*
- * Copyright 2005-2020 Gentoo Foundation
+ * Copyright 2005-2021 Gentoo Foundation
  * Distributed under the terms of the GNU General Public License v2
  *
  * Copyright 2005-2008 Ned Ludd        - <solar@gentoo.org>
@@ -70,14 +70,6 @@ init_coredumps(void)
 }
 #endif
 
-void
-no_colors(void)
-{
-	BOLD = NORM = BLUE = DKBLUE = CYAN = GREEN = DKGREEN = \
-		   MAGENTA = RED = YELLOW = BRYELLOW = WHITE = "";
-	setenv("NOCOLOR", "true", 1);
-}
-
 void
 setup_quiet(void)
 {
@@ -369,6 +361,10 @@ read_portage_file(const char *file, enum portage_file_type type, void *data)
 	char *buf = NULL;
 	size_t buflen = 0;
 	size_t line;
+	bool incomment;
+	size_t cbeg;
+	size_t cend;
+	char npath[_Q_PATH_MAX * 2];
 	int i;
 	env_vars *vars = data;
 	set *masks = data;
@@ -376,7 +372,6 @@ read_portage_file(const char *file, enum portage_file_type type, void *data)
 	if ((dentslen = scandir(file, &dents, NULL, alphasort)) > 0) {
 		int di;
 		struct dirent *d;
-		char npath[_Q_PATH_MAX * 2];
 
 		/* recurse through all files */
 		for (di = 0; di < dentslen; di++) {
@@ -396,17 +391,32 @@ read_portage_file(const char *file, enum portage_file_type type, void *data)
 		goto done;
 
 	line = 0;
+	incomment = false;
+	cbeg = 0;
+	cend = 0;
 	while (getline(&buf, &buflen, fp) != -1) {
 		line++;
 		rmspace(buf);
-		if (*buf == '#' || *buf == '\0')
+		if (*buf == '#') {
+			if (!incomment)
+				cbeg = cend = line;
+			else
+				cend = line;
+			incomment = true;
 			continue;
+		}
+		incomment = false;
+		if (*buf == '\0') {
+			cbeg = cend = 0;
+			continue;
+		}
 
 		/* Handle "source" keyword */
 		if (type == ENV_FILE) {
+			size_t curline = line;
+
 			if (strncmp(buf, "source ", 7) == 0) {
 				const char *sfile = buf + 7;
-				char npath[_Q_PATH_MAX * 2];
 
 				if (sfile[0] != '/') {
 					/* handle relative paths */
@@ -455,6 +465,7 @@ read_portage_file(const char *file, enum portage_file_type type, void *data)
 
 						abuf = NULL;
 						while (getline(&abuf, &abuflen, fp) != -1) {
+							line++;
 							buf = xrealloc(buf, buflen + abuflen);
 							endq = strchr(abuf, q);
 							if (endq)
@@ -483,25 +494,23 @@ read_portage_file(const char *file, enum portage_file_type type, void *data)
 					s[off] = '\0';
 				}
 
-				set_portage_env_var(&vars[i], s, file);
+				snprintf(npath, sizeof(npath), "%s:%zu:%zu-%zu",
+						file, curline, cbeg, cend);
+				set_portage_env_var(&vars[i], s, npath);
 			}
 		} else if (type == PMASK_FILE) {
-			/* trim leading space */
-			for (s = buf; isspace((int)*s); s++)
-				;
-			if (*s == '\0')
-				continue;
-			if (*s == '-') {
+			if (*buf == '-') {
 				/* negation/removal, lookup and drop mask if it exists;
 				 * note that this only supports exact matches (PMS
 				 * 5.2.5) so we don't even have to parse and use
 				 * atom-compare here */
-				s++;
-				if ((p = del_set(s, masks, NULL)) != NULL)
+				if ((p = del_set(buf + 1, masks, NULL)) != NULL)
 					free(p);
 			} else {
-				p = xstrdup(file);
-				if (add_set_value(s, p, masks) != NULL)
+				snprintf(npath, sizeof(npath), "%s:%zu:%zu-%zu",
+						file, line, cbeg, cend);
+				p = xstrdup(npath);
+				if (add_set_value(buf, p, masks) != NULL)
 					free(p);
 			}
 		}
@@ -1024,10 +1033,13 @@ initialize_portage_env(void)
 	if (getenv("PORTAGE_QUIET") != NULL)
 		setup_quiet();
 
-	if (nocolor)
-		no_colors();
-	else
+	if (nocolor) {
+		color_clear();
+		setenv("NOCOLOR", "true", 1);
+	} else {
 		color_remap();
+		unsetenv("NOCOLOR");
+	}
 }
 
 int main(int argc, char **argv)
@@ -1046,11 +1058,11 @@ int main(int argc, char **argv)
 	twidth = 0;
 	if (fstat(fileno(stdout), &st) != -1) {
 		if (!isatty(fileno(stdout))) {
-			no_colors();
+			nocolor = 1;
 		} else {
 			if ((getenv("TERM") == NULL) ||
 					(strcmp(getenv("TERM"), "dumb") == 0))
-				no_colors();
+				nocolor = 1;
 			if (ioctl(0, TIOCGWINSZ, &winsz) == 0 && winsz.ws_col > 0)
 				twidth = (int)winsz.ws_col;
 		}
